import rclpy
from rclpy.action import ActionClient
from rclpy.node import Node

from control_msgs.action import FollowJointTrajectory
from trajectory_msgs.msg import JointTrajectory, JointTrajectoryPoint
from builtin_interfaces.msg import Duration

from hexapod_python.poses import shutdown_pose, interpolate_pose
from hexapod_python.ik import ik

import time
import copy
import functools

class IKNode(Node):
    def __init__(self, node_name = "ik_node"):
        super().__init__(node_name)

        self.leg_groups_ = [ "e1", "e2", "e3", "e4", "e5", "e6"]
        self.legs_active_ = False

        self.joint_names = []
        for leg in self.leg_groups_:
            self.joint_names += [ f"c{leg[1:]}_joint", f"f{leg[1:]}_joint", f"t{leg[1:]}_joint" ] 


        self.action_client_ = ActionClient(self, FollowJointTrajectory, "/joint_group_controller/follow_joint_trajectory")

        self.current_pose_ = copy.deepcopy(shutdown_pose)
        self.move_to_pose(shutdown_pose)

    def is_moving(self):
        return self.legs_active_

    def move_to_pose(self, target_pose, interpolation_steps = 50, modulation_function = {}):
        if(self.is_moving()):
            print("move_to_pose: already moving")
            return
               
        self.follow_poses(interpolate_pose(self.current_pose_, target_pose, interpolation_steps, modulation_function))

    def move_to_poses(self, target_poses, interpolation_steps = 50, modulation_function = {}):
        if(self.is_moving()):
            print("move_to_pose: already moving")
            return

        poses = []
        t = self.current_pose_
        for p in target_poses:
            poses += interpolate_pose(t, p, interpolation_steps, modulation_function)
            t = p

        self.follow_poses(poses)


    def follow_poses(self, poses):
        self.target_pose_ = poses[-1]
        ik_values = [ik(pose) for pose in poses]
        trajectory = self.create_joint_trajectory(ik_values)

        goal = FollowJointTrajectory.Goal(trajectory = trajectory)
        self.action_client_.wait_for_server()

        self.legs_active_ = True
        self.response_future = self.action_client_.send_goal_async(goal)
        self.response_future.add_done_callback(self.response_callback)

    def response_callback(self, future):      
        goal_handle = future.result()
        if goal_handle.accepted:  
            self.result_futures = future.result().get_result_async()        
            self.result_futures.add_done_callback(self.result_callback)

    def result_callback(self, future):
        self.legs_active_ = False
        self.current_pose_ = self.target_pose_


    def create_joint_trajectory(self, ik_values):
        
        t = 0
        joint_trajectory_points = []
        for iv in ik_values:
            joint_trajectory_positions = []
            for leg in self.leg_groups_:
                joint_trajectory_positions += iv[leg]
            
            joint_trajectory_points.append(
                    JointTrajectoryPoint(
                        positions = joint_trajectory_positions,
                        time_from_start = Duration(nanosec = t%50 * 20000000 , sec = int(t/50))
                    )
                )
            t += 1
                    

        return JointTrajectory(
            joint_names = self.joint_names,
            points = joint_trajectory_points
        )


    def publish_ik_values(self, ik_values):
        for leg in self.leg_groups_:
            t = 0
            joint_trajectory_points = []
            for i in ik_values:
                joint_trajectory_points.append(
                    JointTrajectoryPoint(
                        positions = i[leg],
                        time_from_start = Duration(nanosec = t%50 * 20000000 , sec = int(t/50))
                    )
                )
                t += 1


            group_num = leg[1:]
            trajectory = JointTrajectory(
                joint_names = [ f"c{group_num}_joint", f"f{group_num}_joint", f"t{group_num}_joint" ],
                points = joint_trajectory_points
            )
            self.publisher_[leg].publish(trajectory)
