from geometry_msgs.msg import Point, Pose, Quaternion
from hexapod_python.ik import leg_offsets

import copy
import math

standing_pose = {
    'body': Pose(
        position = Point(x = 0.0, y = 0.0, z = 80.0 ),
        orientation = Quaternion(x = 0.0, y = 0.0, z = 0.0, w = 1.0)
    ),
    'legs': {
        'e1': Point(x = leg_offsets['e1'].x +  -50.0, y = leg_offsets['e1'].y + 86.0, z = 0.0),
        'e2': Point(x = leg_offsets['e2'].x + -100.0, y = leg_offsets['e2'].y +  0.0, z = 0.0),
        'e3': Point(x = leg_offsets['e3'].x +  -50.0, y = leg_offsets['e3'].y - 86.0, z = 0.0),
        'e4': Point(x = leg_offsets['e4'].x +   50.0, y = leg_offsets['e4'].y - 86.0, z = 0.0),
        'e5': Point(x = leg_offsets['e5'].x +  100.0, y = leg_offsets['e5'].y +  0.0, z = 0.0),
        'e6': Point(x = leg_offsets['e6'].x +   50.0, y = leg_offsets['e6'].y + 86.0, z = 0.0)
    }
}

stand_up_pose = {
    'body': Pose(
        position = Point(x = 0.0, y = 0.0, z = 80.0 ),
        orientation = Quaternion(x = 0.0, y = 0.0, z = 0.0, w = 1.0)
    ),
    'legs': {
        'e1': Point(x = leg_offsets['e1'].x +  -70.0, y = leg_offsets['e1'].y + 121.0, z = 0.0),
        'e2': Point(x = leg_offsets['e2'].x + -140.0, y = leg_offsets['e2'].y +   0.0, z = 0.0),
        'e3': Point(x = leg_offsets['e3'].x +  -70.0, y = leg_offsets['e3'].y - 121.0, z = 0.0),
        'e4': Point(x = leg_offsets['e4'].x +   70.0, y = leg_offsets['e4'].y - 121.0, z = 0.0),
        'e5': Point(x = leg_offsets['e5'].x +  140.0, y = leg_offsets['e5'].y +   0.0, z = 0.0),
        'e6': Point(x = leg_offsets['e6'].x +   70.0, y = leg_offsets['e6'].y + 121.0, z = 0.0)
    }
}



shutdown_pose = {
    'body': Pose(
        position = Point(x = 0.0, y = 0.0, z = -5.0 ),
        orientation = Quaternion(x = 0.0, y = 0.0, z = 0.0, w = 1.0)
    ),
    'legs': {
        'e1': Point(x = leg_offsets['e1'].x +  -70.0, y = leg_offsets['e1'].y + 121.0, z = 0.0),
        'e2': Point(x = leg_offsets['e2'].x + -140.0, y = leg_offsets['e2'].y +   0.0, z = 0.0),
        'e3': Point(x = leg_offsets['e3'].x +  -70.0, y = leg_offsets['e3'].y - 121.0, z = 0.0),
        'e4': Point(x = leg_offsets['e4'].x +   70.0, y = leg_offsets['e4'].y - 121.0, z = 0.0),
        'e5': Point(x = leg_offsets['e5'].x +  140.0, y = leg_offsets['e5'].y +   0.0, z = 0.0),
        'e6': Point(x = leg_offsets['e6'].x +   70.0, y = leg_offsets['e6'].y + 121.0, z = 0.0)
    }
}

calibration_pose = {
    'body': Pose(
        position = Point(x = 0.0, y = 0.0, z = 0.0 ),
        orientation = Quaternion(x = 0.0, y = 0.0, z = 0.0, w = 1.0)
    ),
    'legs': {
        'e1': Point(x = -200.0, y =  150.0, z = 0.0),
        'e2': Point(x = -200.0, y =    0.0, z = 0.0),
        'e3': Point(x = -200.0, y = -150.0, z = 0.0),         

        'e4': Point(x =  200.0, y = -150.0, z = 0.0),
        'e5': Point(x =  200.0, y =    0.0, z = 0.0),
        'e6': Point(x =  200.0, y =  150.0, z = 0.0),

    }

}


def interpolate_pose(start, end, steps, fn = {}):

    body_step_x = (end['body'].position.x - start['body'].position.x) / (steps-1)
    body_step_y = (end['body'].position.y - start['body'].position.y) / (steps-1)
    body_step_z = (end['body'].position.z - start['body'].position.z) / (steps-1)

    leg_step_x = {}
    leg_step_y = {}
    leg_step_z = {}

    for leg in start['legs']:
        leg_step_x[leg] = (end['legs'][leg].x - start['legs'][leg].x) / (steps-1)
        leg_step_y[leg] = (end['legs'][leg].y - start['legs'][leg].y) / (steps-1)
        leg_step_z[leg] = (end['legs'][leg].z - start['legs'][leg].z) / (steps-1)

    poses = []
    for i in range(0, steps):
        pose = {
            'body': Pose(
                position = Point(
                    x = start['body'].position.x + body_step_x * i, 
                    y = start['body'].position.y + body_step_y * i, 
                    z = start['body'].position.z + body_step_z * i
                ),
                orientation = Quaternion(x = 0.0, y = 0.0, z = 0.0, w = 1.0)
            ),
            'legs': {}            
        }

        for leg in start['legs']:
            pose['legs'][leg] = Point(
                x = start['legs'][leg].x + leg_step_x[leg] * i,
                y = start['legs'][leg].y + leg_step_y[leg] * i,
                z = start['legs'][leg].z + leg_step_z[leg] * i
            )
            if leg in fn:
                (fn_x, fn_y, fn_z) = fn[leg](i)
                pose['legs'][leg].x += fn_x
                pose['legs'][leg].y += fn_y
                pose['legs'][leg].z += fn_z

        poses.append(pose)

    return poses

def step_poses(start_pose, target_pose, legs_at_once, leg_index_offset, interpolation_steps, step_height):
    poses = []
    current_pose = start_pose

    n_legs = len(start_pose['legs'])
    n_cycles = int(n_legs / legs_at_once)

    for cycle in range(0, n_cycles): 
        intermediate_pose = copy.deepcopy(current_pose)
        intermediate_pose['body'].position.x += (target_pose['body'].position.x - start_pose['body'].position.x) / n_cycles
        intermediate_pose['body'].position.y += (target_pose['body'].position.y - start_pose['body'].position.y) / n_cycles
        intermediate_pose['body'].position.z += (target_pose['body'].position.z - start_pose['body'].position.z) / n_cycles
        step_fn = {}
        for l in range(0, legs_at_once): # 0 .. 1
            leg_num = (cycle + 1 + n_cycles * l + leg_index_offset)

            while(leg_num > n_legs):
                    leg_num -= n_legs

            leg = f'e{leg_num}'
            intermediate_pose['legs'][leg] = target_pose['legs'][leg]
            step_fn[leg] = lambda i : (0, 0, step_height * math.sin(i/(interpolation_steps-1) * math.pi))

        poses += interpolate_pose(current_pose, intermediate_pose, interpolation_steps, step_fn)
        current_pose = intermediate_pose

    return poses
