import rclpy
from rclpy.node import Node

from hexapod_python.ik_node import IKNode
from hexapod_python.poses import standing_pose, stand_up_pose, shutdown_pose, interpolate_pose, step_poses

from taranis.msg import TaranisPacket

import math
import time
import copy
import functools



class TaranisTestNode(IKNode):
    MODE_SHUTDOWN = 0
    MODE_STANDING = 1
    MODE_WALKING  = 2

    def __init__(self):
        super().__init__('taranis_test')

        self.subscription = self.create_subscription(
            TaranisPacket,
            '/taranis_output',
            self.taranis_callback,
            10
        )

        self.mode = self.MODE_SHUTDOWN        

    def taranis_callback(self, msg):
        if self.is_moving():
            return
        
        if self.mode == self.MODE_SHUTDOWN:
            if(msg.se == 1):
                self.follow_poses(
                    interpolate_pose(self.current_pose_, stand_up_pose, 50) + 
                    step_poses(stand_up_pose, standing_pose, 2, 0, 25, 20)
                )
                self.mode = self.MODE_STANDING
                print("-> MODE_STANDING")

        if self.mode == self.MODE_STANDING:
            if(msg.se == 0):
                self.follow_poses(
                    step_poses(self.current_pose_, stand_up_pose, 2, 0, 25, 20) +
                    interpolate_pose(stand_up_pose, shutdown_pose, 50) 
                )
                self.mode = self.MODE_SHUTDOWN
                print("-> MODE_SHUTDOWN")

            if(msg.se == 2):
                self.mode = self.MODE_WALKING
                self.second_step = False
                self.last_alpha = 0
                self.last_delta = 0
                self.walking_base_pose = copy.deepcopy(self.current_pose_)
                print("-> MODE_WALKING")

            if(msg.se == 1):
                if(msg.sf == 1):
                    target_pose = copy.deepcopy(self.current_pose_)
                    target_pose['body'].position.x = msg.aileron * 20.0
                    target_pose['body'].position.y = msg.elevator * 20.0
                    target_pose['body'].position.z = standing_pose['body'].position.z + msg.throttle * 20

                    delta = math.sqrt(
                        (target_pose['body'].position.x - self.current_pose_['body'].position.x)**2 +
                        (target_pose['body'].position.y - self.current_pose_['body'].position.y)**2 +
                        (target_pose['body'].position.z - self.current_pose_['body'].position.z)**2
                    )

                    if delta > 0.2:
                        delta_steps = int(delta) * 2
                        if(delta_steps < 2):
                            delta_steps = 2;
                        self.move_to_pose(target_pose, delta_steps)


        if self.mode == self.MODE_WALKING:
            if(msg.se != 2 and self.last_delta == 0):
                self.mode = self.MODE_STANDING
                print("-> MODE_STANDING")
                return

            alpha = -msg.rudder * math.pi / 12 # +/- 30 degrees max

            delta_x = msg.aileron * 100.0 
            delta_y = msg.elevator * 100.0 

            if alpha > 0 and alpha < 0.01:
                alpha = 0
            elif alpha < 0 and alpha > -0.01:
                alpha = 0
                
            delta = math.sqrt(delta_x**2 + delta_y**2)

            if delta < 1:
                delta_x = 0.0
                delta_y = 0.0
                delta = 0


            if delta != 0 or alpha != 0 or self.last_delta != 0 or self.last_alpha != 0:
                step_height = int(50 + 50 * msg.s1)
                interpolation_steps = int(60 + 50 * msg.s2)

                fn = lambda i : (0, 0, step_height * math.sin(i/(interpolation_steps-1) * math.pi))

                lift_leg_group = ['e1', 'e3', 'e5']
                step_pose = copy.deepcopy(self.walking_base_pose)
                step_fn = {}

                for leg in self.leg_groups_:                    
                    x = step_pose['legs'][leg].x
                    y = step_pose['legs'][leg].y

                    if (leg in lift_leg_group) != self.second_step :
                        x += delta_x / 2
                        y += delta_y / 2
                        x = x * math.cos(alpha) - y * math.sin(alpha)
                        y = x * math.sin(alpha) + y + math.cos(alpha)
                        step_fn[leg] = fn
                    else:
                        x -= delta_x / 2
                        y -= delta_y / 2
                        x = x * math.cos(-alpha) - y * math.sin(-alpha)
                        y = x * math.sin(-alpha) + y + math.cos(-alpha)


                    step_pose['legs'][leg].x = x
                    step_pose['legs'][leg].y = y             


                poses = interpolate_pose(self.current_pose_, step_pose, interpolation_steps, step_fn)
                self.follow_poses(poses)
                self.second_step = not self.second_step
                self.last_delta = delta
                self.last_alpha = alpha



            

def main(args = None):
    rclpy.init(args=args)

    ik_publisher = TaranisTestNode()
    rclpy.spin(ik_publisher)
    ik_publisher.destroy_node()
    rclpy.shutdown()
    