#include "taranis/taranis_node.hpp"
#include "rclcpp/rclcpp.hpp"

#include <functional>
#include <memory>
#include <chrono>

using namespace std::chrono_literals;

namespace taranis {

TaranisNode::TaranisNode() : Node("taranis_node") {
    publisher_ = this->create_publisher<taranis::msg::TaranisPacket>("/taranis_output", 10);
    if(!initialize_serial_()) {
        RCLCPP_ERROR(rclcpp::get_logger("TaranisNode"), "serial could not be initialized, exiting.");
        return;
    } else {
        RCLCPP_INFO(rclcpp::get_logger("TaranisNode"), "serial port initialized successfully.");
    }

    timer_ = this->create_wall_timer(
      2ms,
      std::bind(&TaranisNode::timer_callback_, this)
    );
}

bool
TaranisNode::initialize_serial_() {

    serial_fd_ = open("/dev/ttyAMA0", O_RDWR | O_NOCTTY | O_NONBLOCK );
    if (serial_fd_ < 0) {
        RCLCPP_WARN(rclcpp::get_logger("TaranisNode"), "cannot open serial.");
        return false;
    }
    
    if(ioctl(serial_fd_, TCGETS2, &termios2_) < 0) {
        RCLCPP_WARN(rclcpp::get_logger("TaranisNode"), "ioctl TCGETS2 failed.");
        return false;
    }

    // Set up 8-O-2 at 100000 baud, no flow-control, no pi-pa-po

    termios2_.c_iflag &= ~(INPCK | ISTRIP | INLCR | IGNCR | ICRNL | IXON | IXANY | IXOFF);
    termios2_.c_iflag |= IGNBRK | IGNPAR;

    termios2_.c_oflag &= ~(OPOST | ONLCR | OCRNL | ONOCR | ONLRET | OFILL | NLDLY | VTDLY);
    termios2_.c_oflag |= NL0 | VT0;

    termios2_.c_cflag &= ~(CSIZE | CBAUD | (CBAUD << IBSHIFT));
    termios2_.c_cflag |= CS8 | CSTOPB | PARENB | PARODD | CREAD | CLOCAL | BOTHER | (BOTHER << IBSHIFT);

    termios2_.c_lflag &= ~(ISIG | ICANON | ECHO | TOSTOP | IEXTEN);

    termios2_.c_ispeed = 100000;
    termios2_.c_ospeed = 100000;

    termios2_.c_cc[VTIME] = 0;
    termios2_.c_cc[VMIN] = 0;

    if(ioctl(serial_fd_, TCSETS2, &termios2_) < 0) {
        RCLCPP_WARN(rclcpp::get_logger("TaranisNode"), "ioctl TCSETS2 failed.");
        return false;
    }

    return true;
}

void
TaranisNode::timer_callback_() {
    int bytes_read;
    bool did_read = false;

    do {
        
        if(empty_read_ > 2) {
            buffer_index_ = 0;
            empty_read_ = 0;
        }
        
        bytes_read = read(serial_fd_, &buffer_[buffer_index_], 25-buffer_index_);
        
        if(bytes_read) {
            buffer_index_ += bytes_read;
            did_read = true;
            if(buffer_index_ == 25) {
                publish_packet_();
                buffer_index_ = 0;
            }
        } else {
            if(!did_read) empty_read_ ++;
        }
    }
    while(bytes_read);
}

double
TaranisNode::normalize_channel_(uint16_t value) {
    double float_value = value;

    float_value -= 992.0; // (shift center to 0.0)
    float_value /= 820.0;

    if(float_value > 1.0) float_value = 1.0;
    if(float_value < -1.0) float_value = -1.0;

    return float_value;
}

uint8_t
TaranisNode::decode_switch_(uint16_t value) {
    if(value < 582) return 0;
    if(value < 1402) return 1;
    return 2;
}

void
TaranisNode::publish_packet_() {
    if(buffer_[0] != 0x0f || buffer_[24] != 0x00) {
        RCLCPP_WARN(rclcpp::get_logger("TaranisNode"), "invalid sbus frame received, dropping.");
        return;
    }

    uint16_t channels[16];
   
    channels[ 0] = (buffer_[ 2] & 0b00000111) <<  8 |  buffer_[ 1];
    channels[ 1] = (buffer_[ 3] & 0b00111111) <<  5 | (buffer_[ 2] & 0b11111000) >> 3;
    channels[ 2] = (buffer_[ 5] & 0b00000001) << 10 | (buffer_[ 4] & 0b11111111) << 2 | (buffer_[3] & 0b11000000) >> 6;
    channels[ 3] = (buffer_[ 6] & 0b00001111) <<  7 | (buffer_[ 5] & 0b11111110) >> 1;
    channels[ 4] = (buffer_[ 7] & 0b01111111) <<  4 | (buffer_[ 6] & 0b11110000) >> 4;
    channels[ 5] = (buffer_[ 9] & 0b00000011) <<  9 | (buffer_[ 8] & 0b11111111) << 1 | (buffer_[7] & 0b10000000) >> 7;
    channels[ 6] = (buffer_[10] & 0b00011111) <<  6 | (buffer_[ 9] & 0b11111100) >> 2;
    channels[ 7] = (buffer_[11] & 0b11111111) <<  3 | (buffer_[10] & 0b11100000) >> 5;

    channels[ 8] = (buffer_[13] & 0b00000111) <<  8 |  buffer_[12];
    channels[ 9] = (buffer_[14] & 0b00111111) <<  5 | (buffer_[13] & 0b11111000) >> 3;
    channels[10] = (buffer_[16] & 0b00000001) << 10 | (buffer_[15] & 0b11111111) << 2 | (buffer_[14] & 0b11000000) >> 6;
    channels[11] = (buffer_[17] & 0b00001111) <<  7 | (buffer_[16] & 0b11111110) >> 1;
    channels[12] = (buffer_[18] & 0b01111111) <<  4 | (buffer_[17] & 0b11110000) >> 4;
    channels[13] = (buffer_[20] & 0b00000011) <<  9 | (buffer_[19] & 0b11111111) << 1 | (buffer_[18] & 0b10000000) >> 7;
    channels[14] = (buffer_[21] & 0b00011111) <<  6 | (buffer_[20] & 0b11111100) >> 2;
    channels[15] = (buffer_[22] & 0b11111111) <<  3 | (buffer_[21] & 0b11100000) >> 5;

    taranis::msg::TaranisPacket message = taranis::msg::TaranisPacket();

    message.aileron  = normalize_channel_(channels[0]);
    message.elevator = normalize_channel_(channels[1]);
    message.throttle = normalize_channel_(channels[2]);
    message.rudder   = normalize_channel_(channels[3]);

    message.se = decode_switch_(channels[4]);
    message.sf = decode_switch_(channels[5]);
    message.sg = decode_switch_(channels[6]);
    message.sh = decode_switch_(channels[7]);

    message.s1 = normalize_channel_(channels[8]);
    message.s2 = normalize_channel_(channels[9]);

    if(buffer_[23] & (1<<2)) message.signal_loss = true;
    if(buffer_[23] & (1<<3)) message.failsafe_active = true;

    publisher_->publish(message);
}

} // namespace

int main(int argc, char ** argv)
{
  rclcpp::init(argc, argv);
  rclcpp::spin(std::make_shared<taranis::TaranisNode>());
  rclcpp::shutdown();

  return 0;
}
